@@ -5,7 +5,6 @@
 # Released under the same license terms as gPodder itself.
 import functools
 import logging
-import subprocess
 
 import gpodder
 from gpodder import util
@@ -72,7 +71,7 @@ class Win32Player(Player):
             self.command = win32_read_registry_key(self.command)
             return True
         except Exception as e:
-            logger.warn('Win32 player not found: %s (%s)', self.command, e)
+            logger.warning('Win32 player not found: %s (%s)', self.command, e)
 
         return False
 
@@ -212,6 +211,9 @@ PLAYERS = [
     # Clementine, http://www.clementine-player.org/
     FreeDesktopPlayer('clementine', 'Clementine', ['clementine', '--append']),
 
+    # Strawberry, https://www.strawberrymusicplayer.org/
+    FreeDesktopPlayer('strawberry', 'Strawberry', ['strawberry', '--append']),
+
     # Parole, http://docs.xfce.org/apps/parole/start
     FreeDesktopPlayer('parole', 'Parole', ['parole', '-a']),
 
