@@ -24,22 +24,17 @@
 # 2010-04-24 Thomas Perl <thp@gpodder.org>
 #
 
-import gpodder
-_ = gpodder.gettext
-
-import sys
-
+import logging
+import threading
 from sqlite3 import dbapi2 as sqlite
 
-import logging
+import gpodder
+from gpodder import schema, util
+
+_ = gpodder.gettext
+
 logger = logging.getLogger(__name__)
 
-from gpodder import schema
-from gpodder import util
-
-import threading
-import re
-
 
 class Database(object):
     TABLE_PODCAST = 'podcast'
@@ -131,9 +126,12 @@ class Database(object):
         with self.lock:
             cur = self.cursor()
             if podcast_id is not None:
-                cur.execute('SELECT COUNT(*), state, is_new FROM %s WHERE podcast_id = ? GROUP BY state, is_new' % self.TABLE_EPISODE, (podcast_id,))
+                cur.execute('SELECT COUNT(*), state, is_new FROM %s '
+                            'WHERE podcast_id = ? GROUP BY state, is_new'
+                            % self.TABLE_EPISODE, (podcast_id,))
             else:
-                cur.execute('SELECT COUNT(*), state, is_new FROM %s GROUP BY state, is_new' % self.TABLE_EPISODE)
+                cur.execute('SELECT COUNT(*), state, is_new FROM %s '
+                            'GROUP BY state, is_new' % self.TABLE_EPISODE)
             for count, state, is_new in cur:
                 total += count
                 if state == gpodder.STATE_DELETED:
