@@ -25,20 +25,22 @@
 #  Based on libwget.py (2005-10-29)
 #
 
-import collections
-import email
+import glob
 import logging
 import mimetypes
 import os
 import os.path
 import shutil
-import socket
 import threading
 import time
 import urllib.error
-import urllib.parse
-import urllib.request
-from email.header import decode_header
+from abc import ABC, abstractmethod
+
+import requests
+from requests.adapters import HTTPAdapter
+from requests.exceptions import ConnectionError, HTTPError, RequestException
+from requests.packages.urllib3.exceptions import MaxRetryError
+from requests.packages.urllib3.util.retry import Retry
 
 import gpodder
 from gpodder import registry, util
@@ -47,10 +49,27 @@ logger = logging.getLogger(__name__)
 
 _ = gpodder.gettext
 
+REDIRECT_RETRIES = 3
 
-class CustomDownload:
+
+class CustomDownload(ABC):
     """ abstract class for custom downloads. DownloadTask call retrieve_resume() on it """
 
+    @property
+    @abstractmethod
+    def partial_filename(self):
+        """
+        Full path to the temporary file actually being downloaded (downloaders
+        may not support setting a tempname).
+        """
+        ...
+
+    @partial_filename.setter
+    @abstractmethod
+    def partial_filename(self, val):
+        ...
+
+    @abstractmethod
     def retrieve_resume(self, tempname, reporthook):
         """
         :param str tempname: temporary filename for the download
@@ -60,13 +79,14 @@ class CustomDownload:
         return {}, None
 
 
-class CustomDownloader:
+class CustomDownloader(ABC):
     """
     abstract class for custom downloaders.
 
     DownloadTask calls custom_downloader to get a CustomDownload
     """
 
+    @abstractmethod
     def custom_downloader(self, config, episode):
         """
         if this custom downloader has a custom download method (e.g. youtube-dl),
@@ -78,30 +98,6 @@ class CustomDownloader:
         return None
 
 
-def get_header_param(headers, param, header_name):
-    """Extract a HTTP header parameter from a dict
-
-    Uses the "email" module to retrieve parameters
-    from HTTP headers. This can be used to get the
-    "filename" parameter of the "content-disposition"
-    header for downloads to pick a good filename.
-
-    Returns None if the filename cannot be retrieved.
-    """
-    value = None
-    try:
-        headers_string = ['%s:%s' % (k, v) for k, v in list(headers.items())]
-        msg = email.message_from_string('\n'.join(headers_string))
-        if header_name in msg:
-            raw_value = msg.get_param(param, header=header_name)
-            if raw_value is not None:
-                value = email.utils.collapse_rfc2231_value(raw_value)
-    except Exception as e:
-        logger.error('Cannot get %s from %s', param, header_name, exc_info=True)
-
-    return value
-
-
 class ContentRange(object):
     # Based on:
     # http://svn.pythonpaste.org/Paste/WebOb/trunk/webob/byterange.py
@@ -207,7 +203,7 @@ class ContentRange(object):
 class DownloadCancelledException(Exception): pass
 
 
-class AuthenticationError(Exception): pass
+class DownloadNoURLException(Exception): pass
 
 
 class gPodderDownloadHTTPError(Exception):
@@ -217,75 +213,39 @@ class gPodderDownloadHTTPError(Exception):
         self.error_message = error_message
 
 
-class DownloadURLOpener(urllib.request.FancyURLopener):
-    version = gpodder.user_agent
+class DownloadURLOpener:
 
     # Sometimes URLs are not escaped correctly - try to fix them
     # (see RFC2396; Section 2.4.3. Excluded US-ASCII Characters)
     # FYI: The omission of "%" in the list is to avoid double escaping!
     ESCAPE_CHARS = dict((ord(c), '%%%x' % ord(c)) for c in ' <>#"{}|\\^[]`')
 
-    def __init__(self, channel):
-        self.channel = channel
-        self._auth_retry_counter = 0
+    def __init__(self, channel, max_retries=3):
         super().__init__()
+        self.channel = channel
+        self.max_retries = max_retries
 
-    def http_error(self, url, fp, errcode, errmsg, headers, data=None):
-        """Handle http errors.
-        Overriden to give retry=True to http_error_40{1,7}.
-        See https://github.com/python/cpython/commit/80f1b059714aeb1c6fc9f6ce1173bc8a51af7dd9
-        See python issue https://bugs.python.org/issue1368368
-        """
-        result = False
-        if errcode == 401:
-            result = self.http_error_401(url, fp, errcode, errmsg, headers, data=data, retry=True)
-        elif errcode == 407:
-            result = self.http_error_407(url, fp, errcode, errmsg, headers, data=data, retry=True)
-        if result:
-            return result
-        return super().http_error(url, fp, errcode, errmsg, headers, data=data)
-
-    def http_error_default(self, url, fp, errcode, errmsg, headers):
-        """
-        FancyURLopener by default does not raise an exception when
-        there is some unknown HTTP error code. We want to override
-        this and provide a function to log the error and raise an
-        exception, so we don't download the HTTP error page here.
-        """
-        # The following two lines are copied from urllib.URLopener's
-        # implementation of http_error_default
-        void = fp.read()
-        fp.close()
-        raise gPodderDownloadHTTPError(url, errcode, errmsg)
-
-    def redirect_internal(self, url, fp, errcode, errmsg, headers, data):
-        """ This is the exact same function that's included with urllib
-            except with "void = fp.read()" commented out. """
-
-        if 'location' in headers:
-            newurl = headers['location']
-        elif 'uri' in headers:
-            newurl = headers['uri']
-        else:
-            return
-
-        # This blocks forever(?) with certain servers (see bug #465)
-        # void = fp.read()
-        fp.close()
-
-        # In case the server sent a relative URL, join with original:
-        newurl = urllib.parse.urljoin(self.type + ":" + url, newurl)
-        return self.open(newurl)
+    def init_session(self):
+        """ init a session with our own retry codes + retry count """
+        # I add a few retries for redirects but it means that I will allow max_retries + REDIRECT_RETRIES
+        # if encountering max_retries connect and REDIRECT_RETRIES read for instance
+        retry_strategy = Retry(
+            total=self.max_retries + REDIRECT_RETRIES,
+            connect=self.max_retries,
+            read=self.max_retries,
+            redirect=max(REDIRECT_RETRIES, self.max_retries),
+            status=self.max_retries,
+            status_forcelist=Retry.RETRY_AFTER_STATUS_CODES.union((408, 418, 504, 598, 599,)))
+        adapter = HTTPAdapter(max_retries=retry_strategy)
+        http = requests.Session()
+        http.mount("https://", adapter)
+        http.mount("http://", adapter)
+        return http
 
 # The following is based on Python's urllib.py "URLopener.retrieve"
 # Also based on http://mail.python.org/pipermail/python-list/2001-October/110069.html
 
-    def http_error_206(self, url, fp, errcode, errmsg, headers, data=None):
-        # The next line is taken from urllib's URLopener.open_http
-        # method, at the end after the line "if errcode == 200:"
-        return urllib.addinfourl(fp, headers, 'http:' + url)
-
-    def retrieve_resume(self, url, filename, reporthook=None, data=None):
+    def retrieve_resume(self, url, filename, reporthook=None, data=None, disable_auth=False):
         """Download files from an URL; return (headers, real_url)
 
         Resumes a download if the local filename exists and
@@ -294,15 +254,25 @@ class DownloadURLOpener(urllib.request.FancyURLopener):
 
         current_size = 0
         tfp = None
+        headers = {
+            'User-agent': gpodder.user_agent
+        }
+
+        if (self.channel.auth_username or self.channel.auth_password) and not disable_auth:
+            logger.debug('Authenticating as "%s"', self.channel.auth_username)
+            auth = (self.channel.auth_username, self.channel.auth_password)
+        else:
+            auth = None
+
         if os.path.exists(filename):
             try:
                 current_size = os.path.getsize(filename)
                 tfp = open(filename, 'ab')
                 # If the file exists, then only download the remainder
                 if current_size > 0:
-                    self.addheader('Range', 'bytes=%s-' % (current_size))
+                    headers['Range'] = 'bytes=%s-' % (current_size)
             except:
-                logger.warn('Cannot resume download: %s', filename, exc_info=True)
+                logger.warning('Cannot resume download: %s', filename, exc_info=True)
                 tfp = None
                 current_size = 0
 
@@ -312,47 +282,53 @@ class DownloadURLOpener(urllib.request.FancyURLopener):
         # Fix a problem with bad URLs that are not encoded correctly (bug 549)
         url = url.translate(self.ESCAPE_CHARS)
 
-        fp = self.open(url, data)
-        headers = fp.info()
+        session = self.init_session()
+        with session.get(url,
+                         headers=headers,
+                         stream=True,
+                         auth=auth,
+                         timeout=gpodder.SOCKET_TIMEOUT) as resp:
+            try:
+                resp.raise_for_status()
+            except HTTPError as e:
+                if auth is not None:
+                    # Try again without authentication (bug 1296)
+                    return self.retrieve_resume(url, filename, reporthook, data, True)
+                else:
+                    raise gPodderDownloadHTTPError(url, resp.status_code, str(e))
 
-        if current_size > 0:
-            # We told the server to resume - see if she agrees
-            # See RFC2616 (206 Partial Content + Section 14.16)
-            # XXX check status code here, too...
-            range = ContentRange.parse(headers.get('content-range', ''))
-            if range is None or range.start != current_size:
-                # Ok, that did not work. Reset the download
-                # TODO: seek and truncate if content-range differs from request
-                tfp.close()
-                tfp = open(filename, 'wb')
-                current_size = 0
-                logger.warn('Cannot resume: Invalid Content-Range (RFC2616).')
+            headers = resp.headers
 
-        result = headers, fp.geturl()
-        bs = 1024 * 8
-        size = -1
-        read = current_size
-        blocknum = current_size // bs
-        if reporthook:
-            if "content-length" in headers:
-                size = int(headers['Content-Length']) + current_size
-            reporthook(blocknum, bs, size)
-        while read < size or size == -1:
-            if size == -1:
-                block = fp.read(bs)
-            else:
-                block = fp.read(min(size - read, bs))
-            if len(block) == 0:
-                break
-            read += len(block)
-            tfp.write(block)
-            blocknum += 1
+            if current_size > 0:
+                # We told the server to resume - see if she agrees
+                # See RFC2616 (206 Partial Content + Section 14.16)
+                # XXX check status code here, too...
+                range = ContentRange.parse(headers.get('content-range', ''))
+                if range is None or range.start != current_size:
+                    # Ok, that did not work. Reset the download
+                    # TODO: seek and truncate if content-range differs from request
+                    tfp.close()
+                    tfp = open(filename, 'wb')
+                    current_size = 0
+                    logger.warning('Cannot resume: Invalid Content-Range (RFC2616).')
+
+            result = headers, resp.url
+            bs = 1024 * 8
+            size = -1
+            read = current_size
+            blocknum = current_size // bs
             if reporthook:
+                if "content-length" in headers:
+                    size = int(headers['content-length']) + current_size
                 reporthook(blocknum, bs, size)
-        fp.close()
-        tfp.close()
-        del fp
-        del tfp
+            for block in resp.iter_content(bs):
+                read += len(block)
+                tfp.write(block)
+                blocknum += 1
+                if reporthook:
+                    reporthook(blocknum, bs, size)
+            tfp.close()
+            del tfp
 
         # raise exception if actual size does not match content-length header
         if size >= 0 and read < size:
@@ -363,36 +339,30 @@ class DownloadURLOpener(urllib.request.FancyURLopener):
 
 # end code based on urllib.py
 
-    def prompt_user_passwd(self, host, realm):
-        # Keep track of authentication attempts, fail after the third one
-        self._auth_retry_counter += 1
-        if self._auth_retry_counter > 3:
-            raise AuthenticationError(_('Wrong username/password'))
-
-        if self.channel.auth_username or self.channel.auth_password:
-            logger.debug('Authenticating as "%s" to "%s" for realm "%s".',
-                    self.channel.auth_username, host, realm)
-            return (self.channel.auth_username, self.channel.auth_password)
-
-        return (None, None)
-
 
 class DefaultDownload(CustomDownload):
     def __init__(self, config, episode, url):
         self._config = config
         self.__episode = episode
         self._url = url
+        self.__partial_filename = None
+
+    @property
+    def partial_filename(self):
+        return self.__partial_filename
+
+    @partial_filename.setter
+    def partial_filename(self, val):
+        self.__partial_filename = val
 
     def retrieve_resume(self, tempname, reporthook):
         url = self._url
         logger.info("Downloading %s", url)
-        downloader = DownloadURLOpener(self.__episode.channel)
-
-        # HTTP Status codes for which we retry the download
-        retry_codes = (408, 418, 504, 598, 599)
         max_retries = max(0, self._config.auto.retries)
+        downloader = DownloadURLOpener(self.__episode.channel, max_retries=max_retries)
+        self.partial_filename = tempname
 
-        # Retry the download on timeout (bug 1013)
+        # Retry the download on incomplete download (other retries are done by the Retry strategy)
         for retry in range(max_retries + 1):
             if retry > 0:
                 logger.info('Retrying download of %s (%d)', url, retry)
@@ -409,17 +379,6 @@ class DefaultDownload(CustomDownload):
                             url)
                     continue
                 raise
-            except socket.timeout as tmout:
-                if retry < max_retries:
-                    logger.info('Socket timeout: %s - will retry.', url)
-                    continue
-                raise
-            except gPodderDownloadHTTPError as http:
-                if retry < max_retries and http.error_code in retry_codes:
-                    logger.info('HTTP error %d: %s - will retry.',
-                            http.error_code, url)
-                    continue
-                raise
         return (headers, real_url)
 
 
@@ -453,14 +412,14 @@ class DownloadQueueWorker(object):
             if not self.continue_check_callback(self):
                 return
 
-            try:
-                task = self.queue.get_next()
-                logger.info('%s is processing: %s', self, task)
-                task.run()
-                task.recycle()
-            except StopIteration as e:
+            task = self.queue.get_next() if self.queue.enabled else None
+            if not task:
                 logger.info('No more tasks for %s to carry out.', self)
                 break
+            logger.info('%s is processing: %s', self, task)
+            task.run()
+            task.recycle()
+
         self.exit_callback(self)
 
 
@@ -475,6 +434,7 @@ class ForceDownloadWorker(object):
         logger.info('Starting new thread: %s', self)
         logger.info('%s is processing: %s', self, self.task)
         self.task.run()
+        self.task.recycle()
 
 
 class DownloadQueueManager(object):
@@ -485,14 +445,21 @@ class DownloadQueueManager(object):
         self.worker_threads_access = threading.RLock()
         self.worker_threads = []
 
+    def disable(self):
+        self.tasks.enabled = False
+
+    def enable(self):
+        self.tasks.enabled = True
+        self.__spawn_threads()
+
     def __exit_callback(self, worker_thread):
         with self.worker_threads_access:
             self.worker_threads.remove(worker_thread)
 
     def __continue_check_callback(self, worker_thread):
         with self.worker_threads_access:
-            if len(self.worker_threads) > self._config.max_downloads and \
-                    self._config.max_downloads_enabled:
+            if len(self.worker_threads) > self._config.limit.downloads.concurrent and \
+                    self._config.limit.downloads.enabled:
                 self.worker_threads.remove(worker_thread)
                 return False
             else:
@@ -501,16 +468,19 @@ class DownloadQueueManager(object):
     def __spawn_threads(self):
         """Spawn new worker threads if necessary
         """
+        if not self.tasks.enabled:
+            return
+
         with self.worker_threads_access:
             work_count = self.tasks.available_work_count()
-            if self._config.max_downloads_enabled:
+            if self._config.limit.downloads.enabled:
                 # always allow at least 1 download
-                max_downloads = max(int(self._config.max_downloads), 1)
-                spawn_limit = max_downloads - len(self.worker_threads)
+                spawn_limit = max(int(self._config.limit.downloads.concurrent), 1)
             else:
                 spawn_limit = self._config.limit.downloads.concurrent_max
-            logger.info('%r tasks to do, can start at most %r threads', work_count, spawn_limit)
-            for i in range(0, min(work_count, spawn_limit)):
+            running = len(self.worker_threads)
+            logger.info('%r tasks to do, can start at most %r threads, %r threads currently running', work_count, spawn_limit, running)
+            for i in range(0, min(work_count, spawn_limit - running)):
                 # We have to create a new thread here, there's work to do
                 logger.info('Starting new worker thread.')
 
@@ -523,16 +493,21 @@ class DownloadQueueManager(object):
         self.__spawn_threads()
 
     def force_start_task(self, task):
-        if self.tasks.set_downloading(task):
-            worker = ForceDownloadWorker(task)
-            util.run_in_background(worker.run)
+        with task:
+            if task.status in (task.QUEUED, task.PAUSED, task.CANCELLED, task.FAILED):
+                task.status = task.DOWNLOADING
+                worker = ForceDownloadWorker(task)
+                util.run_in_background(worker.run)
 
     def queue_task(self, task):
         """Marks a task as queued
         """
-        task.status = DownloadTask.QUEUED
+        self.tasks.queue_task(task)
         self.__spawn_threads()
 
+    def has_workers(self):
+        return len(self.worker_threads) > 0
+
 
 class DownloadTask(object):
     """An object representing the download task of an episode
@@ -540,7 +515,7 @@ class DownloadTask(object):
     You can create a new download task like this:
 
         task = DownloadTask(episode, gpodder.config.Config(CONFIGFILE))
-        task.status = DownloadTask.DOWNLOADING
+        task.status = DownloadTask.QUEUED
         task.run()
 
     While the download is in progress, you can access its properties:
@@ -557,7 +532,8 @@ class DownloadTask(object):
 
     You can cancel a running download task by setting its status:
 
-        task.status = DownloadTask.CANCELLED
+        with task:
+            task.status = DownloadTask.CANCELLING
 
     The task will then abort as soon as possible (due to the nature
     of downloading data, this can take a while when the Internet is
@@ -608,9 +584,9 @@ class DownloadTask(object):
     The same thing works for failed downloads ("notify_as_failed()").
     """
     # Possible states this download task can be in
-    STATUS_MESSAGE = (_('Added'), _('Queued'), _('Downloading'),
-            _('Finished'), _('Failed'), _('Cancelled'), _('Paused'))
-    (INIT, QUEUED, DOWNLOADING, DONE, FAILED, CANCELLED, PAUSED) = list(range(7))
+    STATUS_MESSAGE = (_('Queued'), _('Queued'), _('Downloading'),
+            _('Finished'), _('Failed'), _('Cancelling'), _('Cancelled'), _('Pausing'), _('Paused'))
+    (NEW, QUEUED, DOWNLOADING, DONE, FAILED, CANCELLING, CANCELLED, PAUSING, PAUSED) = list(range(9))
 
     # Wheter this task represents a file download or a device sync operation
     ACTIVITY_DOWNLOAD, ACTIVITY_SYNCHRONIZE = list(range(2))
@@ -621,6 +597,12 @@ class DownloadTask(object):
     def __str__(self):
         return self.__episode.title
 
+    def __enter__(self):
+        return self.__lock.acquire()
+
+    def __exit__(self, type, value, traceback):
+        self.__lock.release()
+
     def __get_status(self):
         return self.__status
 
@@ -672,17 +654,61 @@ class DownloadTask(object):
 
     downloader = property(fget=__get_downloader, fset=__set_downloader)
 
+    def can_queue(self):
+        return self.status in (self.CANCELLED, self.PAUSED, self.FAILED)
+
+    def unpause(self):
+        with self:
+            # Resume a downloading task that was transitioning to paused
+            if self.status == self.PAUSING:
+                self.status = self.DOWNLOADING
+
+    def can_pause(self):
+        return self.status in (self.DOWNLOADING, self.QUEUED)
+
+    def pause(self):
+        with self:
+            # Pause a queued download
+            if self.status == self.QUEUED:
+                self.status = self.PAUSED
+            # Request pause of a running download
+            elif self.status == self.DOWNLOADING:
+                self.status = self.PAUSING
+                # download rate limited tasks sleep and take longer to transition from the PAUSING state to the PAUSED state
+
+    def can_cancel(self):
+        return self.status in (self.DOWNLOADING, self.QUEUED, self.PAUSED, self.FAILED)
+
     def cancel(self):
-        if self.status in (self.DOWNLOADING, self.QUEUED):
-            self.status = self.CANCELLED
+        with self:
+            # Cancelling directly is allowed if the task isn't currently downloading
+            if self.status in (self.QUEUED, self.PAUSED, self.FAILED):
+                self.status = self.CANCELLING
+                # Call run, so the partial file gets deleted, and task recycled
+                self.run()
+            # Otherwise request cancellation
+            elif self.status == self.DOWNLOADING:
+                self.status = self.CANCELLING
+
+    def can_remove(self):
+        return self.status in (self.CANCELLED, self.FAILED, self.DONE)
+
+    def delete_partial_files(self):
+        temporary_files = [self.tempname]
+        # youtube-dl creates .partial.* files for adaptive formats
+        temporary_files += glob.glob('%s.*' % self.tempname)
+
+        for tempfile in temporary_files:
+            util.delete_file(tempfile)
 
     def removed_from_list(self):
         if self.status != self.DONE:
-            util.delete_file(self.tempname)
+            self.delete_partial_files()
 
     def __init__(self, episode, config, downloader=None):
         assert episode.download_task is None
-        self.__status = DownloadTask.INIT
+        self.__lock = threading.RLock()
+        self.__status = DownloadTask.NEW
         self.__activity = DownloadTask.ACTIVITY_DOWNLOAD
         self.__status_changed = True
         self.__episode = episode
@@ -698,6 +724,7 @@ class DownloadTask(object):
         self.speed = 0.0
         self.progress = 0.0
         self.error_message = None
+        self.custom_downloader = None
 
         # Have we already shown this task in a notification?
         self._notification_shown = False
@@ -705,8 +732,8 @@ class DownloadTask(object):
         # Variables for speed limit and speed calculation
         self.__start_time = 0
         self.__start_blocks = 0
-        self.__limit_rate_value = self._config.limit_rate_value
-        self.__limit_rate = self._config.limit_rate
+        self.__limit_rate_value = self._config.limit.bandwidth.kbps
+        self.__limit_rate = self._config.limit.bandwidth.enabled
 
         # Progress update functions
         self._progress_updated = None
@@ -728,6 +755,11 @@ class DownloadTask(object):
         # Store a reference to this task in the episode
         episode.download_task = self
 
+    def reuse(self):
+        if not os.path.exists(self.tempname):
+            # partial file was deleted when cancelled, recreate it
+            open(self.tempname, 'w').close()
+
     def notify_as_finished(self):
         if self.status == DownloadTask.DONE:
             if self._notification_shown:
@@ -772,10 +804,10 @@ class DownloadTask(object):
 
         self.calculate_speed(count, blockSize)
 
-        if self.status == DownloadTask.CANCELLED:
+        if self.status == DownloadTask.CANCELLING:
             raise DownloadCancelledException()
 
-        if self.status == DownloadTask.PAUSED:
+        if self.status == DownloadTask.PAUSING:
             raise DownloadCancelledException()
 
     def calculate_speed(self, count, blockSize):
@@ -783,18 +815,18 @@ class DownloadTask(object):
             now = time.time()
             if self.__start_time > 0:
                 # Has rate limiting been enabled or disabled?
-                if self.__limit_rate != self._config.limit_rate:
+                if self.__limit_rate != self._config.limit.bandwidth.enabled:
                     # If it has been enabled then reset base time and block count
-                    if self._config.limit_rate:
+                    if self._config.limit.bandwidth.enabled:
                         self.__start_time = now
                         self.__start_blocks = count
-                    self.__limit_rate = self._config.limit_rate
+                    self.__limit_rate = self._config.limit.bandwidth.enabled
 
                 # Has the rate been changed and are we currently limiting?
-                if self.__limit_rate_value != self._config.limit_rate_value and self.__limit_rate:
+                if self.__limit_rate_value != self._config.limit.bandwidth.kbps and self.__limit_rate:
                     self.__start_time = now
                     self.__start_blocks = count
-                    self.__limit_rate_value = self._config.limit_rate_value
+                    self.__limit_rate_value = self._config.limit.bandwidth.kbps
 
                 passed = now - self.__start_time
                 if passed > 0:
@@ -809,47 +841,61 @@ class DownloadTask(object):
 
             self.speed = float(speed)
 
-            if self._config.limit_rate and speed > self._config.limit_rate_value:
+            if self._config.limit.bandwidth.enabled and speed > self._config.limit.bandwidth.kbps:
                 # calculate the time that should have passed to reach
                 # the desired download rate and wait if necessary
-                should_have_passed = (count - self.__start_blocks) * blockSize / (self._config.limit_rate_value * 1024.0)
+                should_have_passed = (count - self.__start_blocks) * blockSize / (self._config.limit.bandwidth.kbps * 1024.0)
                 if should_have_passed > passed:
                     # sleep a maximum of 10 seconds to not cause time-outs
                     delay = min(10.0, float(should_have_passed - passed))
                     time.sleep(delay)
 
     def recycle(self):
-        self.episode.download_task = None
+        if self.status not in (self.FAILED, self.PAUSED):
+            self.episode.download_task = None
+
+    def set_episode_download_task(self):
+        if not self.episode.download_task:
+            self.episode.download_task = self
 
     def run(self):
         # Speed calculation (re-)starts here
         self.__start_time = 0
         self.__start_blocks = 0
 
-        # If the download has already been cancelled, skip it
-        if self.status == DownloadTask.CANCELLED:
-            util.delete_file(self.tempname)
-            self.progress = 0.0
-            self.speed = 0.0
-            self.recycle()
-            return False
+        # If the download has already been cancelled/paused, skip it
+        with self:
+            if self.status == DownloadTask.CANCELLING:
+                self.status = DownloadTask.CANCELLED
+                self.__episode._download_error = None
+                self.delete_partial_files()
+                self.progress = 0.0
+                self.speed = 0.0
+                self.recycle()
+                return False
 
-        # We only start this download if its status is "downloading"
-        if self.status != DownloadTask.DOWNLOADING:
-            return False
+            if self.status == DownloadTask.PAUSING:
+                self.status = DownloadTask.PAUSED
+                return False
 
-        # We are downloading this file right now
-        self.status = DownloadTask.DOWNLOADING
-        self._notification_shown = False
+            # We only start this download if its status is downloading
+            if self.status != DownloadTask.DOWNLOADING:
+                return False
 
-        # Restore a reference to this task in the episode
-        # when running a recycled task following a pause or failed
-        # see #649
-        if not self.episode.download_task:
-            self.episode.download_task = self
+            # We are downloading this file right now
+            self._notification_shown = False
+
+            # Restore a reference to this task in the episode
+            # when running a recycled task following a pause or failed
+            # see #649
+            self.set_episode_download_task()
 
         url = self.__episode.url
+        result = DownloadTask.DOWNLOADING
         try:
+            if url == '':
+                raise DownloadNoURLException()
+
             if self.downloader:
                 downloader = self.downloader.custom_downloader(self._config, self.episode)
             else:
@@ -860,6 +906,7 @@ class DownloadTask(object):
             else:
                 downloader = DefaultDownloader.custom_downloader(self._config, self.episode)
 
+            self.custom_downloader = downloader
             headers, real_url = downloader.retrieve_resume(self.tempname, self.status_updated)
 
             new_mimetype = headers.get('content-type', self.__episode.mime_type)
@@ -893,7 +940,7 @@ class DownloadTask(object):
                             real_url, os.path.basename(self.filename))
 
             # Look at the Content-disposition header; use if if available
-            disposition_filename = get_header_param(headers, 'filename', 'content-disposition')
+            disposition_filename = util.get_header_param(headers, 'filename', 'content-disposition')
 
             # Some servers do send the content-disposition header, but provide
             # an empty filename, resulting in an empty string here (bug 1440)
@@ -922,42 +969,73 @@ class DownloadTask(object):
             self.__episode.on_downloaded(self.filename)
         except DownloadCancelledException:
             logger.info('Download has been cancelled/paused: %s', self)
-            if self.status == DownloadTask.CANCELLED:
-                util.delete_file(self.tempname)
+            if self.status == DownloadTask.CANCELLING:
+                self.__episode._download_error = None
+                self.delete_partial_files()
                 self.progress = 0.0
                 self.speed = 0.0
+            result = DownloadTask.CANCELLED
+        except DownloadNoURLException:
+            result = DownloadTask.FAILED
+            self.error_message = _('Episode has no URL to download')
         except urllib.error.ContentTooShortError as ctse:
-            self.status = DownloadTask.FAILED
+            result = DownloadTask.FAILED
             self.error_message = _('Missing content from server')
+        except ConnectionError as ce:
+            # special case request exception
+            result = DownloadTask.FAILED
+            logger.error('Download failed: %s', str(ce))
+            d = {'host': ce.args[0].pool.host, 'port': ce.args[0].pool.port}
+            self.error_message = _("Couldn't connect to server %(host)s:%(port)s" % d)
+        except RequestException as re:
+            # extract MaxRetryError to shorten the exception message
+            if isinstance(re.args[0], MaxRetryError):
+                re = re.args[0]
+            logger.error('%s while downloading "%s"', str(re),
+                    self.__episode.title)
+            result = DownloadTask.FAILED
+            d = {'error': str(re)}
+            self.error_message = _('Request Error: %(error)s') % d
         except IOError as ioe:
             logger.error('%s while downloading "%s": %s', ioe.strerror,
-                    self.__episode.title, ioe.filename, exc_info=True)
-            self.status = DownloadTask.FAILED
+                    self.__episode.title, ioe.filename)
+            result = DownloadTask.FAILED
             d = {'error': ioe.strerror, 'filename': ioe.filename}
             self.error_message = _('I/O Error: %(error)s: %(filename)s') % d
         except gPodderDownloadHTTPError as gdhe:
             logger.error('HTTP %s while downloading "%s": %s',
-                    gdhe.error_code, self.__episode.title, gdhe.error_message,
-                    exc_info=True)
-            self.status = DownloadTask.FAILED
+                    gdhe.error_code, self.__episode.title, gdhe.error_message)
+            result = DownloadTask.FAILED
             d = {'code': gdhe.error_code, 'message': gdhe.error_message}
             self.error_message = _('HTTP Error %(code)s: %(message)s') % d
         except Exception as e:
-            self.status = DownloadTask.FAILED
+            result = DownloadTask.FAILED
             logger.error('Download failed: %s', str(e), exc_info=True)
             self.error_message = _('Error: %s') % (str(e),)
 
-        if self.status == DownloadTask.DOWNLOADING:
-            # Everything went well - we're done
-            self.status = DownloadTask.DONE
-            if self.total_size <= 0:
-                self.total_size = util.calculate_size(self.filename)
-                logger.info('Total size updated to %d', self.total_size)
-            self.progress = 1.0
-            gpodder.user_extensions.on_episode_downloaded(self.__episode)
-            return True
+        with self:
+            if result == DownloadTask.DOWNLOADING:
+                # Everything went well - we're done (even if the task was cancelled/paused,
+                # since it's finished we might as well mark it done)
+                self.status = DownloadTask.DONE
+                if self.total_size <= 0:
+                    self.total_size = util.calculate_size(self.filename)
+                    logger.info('Total size updated to %d', self.total_size)
+                self.progress = 1.0
+                gpodder.user_extensions.on_episode_downloaded(self.__episode)
+                return True
 
-        self.speed = 0.0
+            self.speed = 0.0
+
+            if result == DownloadTask.FAILED:
+                self.status = DownloadTask.FAILED
+                self.__episode._download_error = self.error_message
+
+            # cancelled/paused -- update state to mark it as safe to manipulate this task again
+            elif self.status == DownloadTask.PAUSING:
+                self.status = DownloadTask.PAUSED
+            elif self.status == DownloadTask.CANCELLING:
+                self.status = DownloadTask.CANCELLED
 
         # We finished, but not successfully (at least not really)
         return False
