@@ -19,14 +19,14 @@
 
 import html
 import logging
-import urllib.parse
 
 from gi.repository import Gdk, Gtk, Pango
 
 import gpodder
 from gpodder import util, vimeo, youtube
 from gpodder.gtkui.desktopfile import PlayerListModel
-from gpodder.gtkui.interface.common import BuilderWidget, TreeViewHelper
+from gpodder.gtkui.interface.common import (BuilderWidget, TreeViewHelper,
+                                            show_message_dialog)
 from gpodder.gtkui.interface.configeditor import gPodderConfigEditor
 
 logger = logging.getLogger(__name__)
@@ -50,13 +50,13 @@ class NewEpisodeActionList(Gtk.ListStore):
 
     def get_index(self):
         for index, row in enumerate(self):
-            if self._config.auto_download == row[self.C_AUTO_DOWNLOAD]:
+            if self._config.ui.gtk.new_episodes == row[self.C_AUTO_DOWNLOAD]:
                 return index
 
         return 1  # Some sane default
 
     def set_index(self, index):
-        self._config.auto_download = self[index][self.C_AUTO_DOWNLOAD]
+        self._config.ui.gtk.new_episodes = self[index][self.C_AUTO_DOWNLOAD]
 
 
 class DeviceTypeActionList(Gtk.ListStore):
@@ -92,11 +92,11 @@ class OnSyncActionList(Gtk.ListStore):
 
     def get_index(self):
         for index, row in enumerate(self):
-            if (self._config.device_sync.after_sync.delete_episodes and
-                    row[self.C_ON_SYNC_DELETE]):
+            if (self._config.device_sync.after_sync.delete_episodes
+                    and row[self.C_ON_SYNC_DELETE]):
                 return index
-            if (self._config.device_sync.after_sync.mark_episodes_played and
-                    row[self.C_ON_SYNC_MARK_PLAYED] and not
+            if (self._config.device_sync.after_sync.mark_episodes_played
+                    and row[self.C_ON_SYNC_MARK_PLAYED] and not
                     self._config.device_sync.after_sync.delete_episodes):
                 return index
         return 0  # Some sane default
@@ -184,6 +184,7 @@ class gPodderPreferences(BuilderWidget):
     C_TOGGLE, C_LABEL, C_EXTENSION, C_SHOW_TOGGLE = list(range(4))
 
     def new(self):
+        self.gPodderPreferences.set_transient_for(self.parent_widget)
         for cb in (self.combo_audio_player_app, self.combo_video_player_app):
             cellrenderer = Gtk.CellRendererPixbuf()
             cb.pack_start(cellrenderer, False)
@@ -196,17 +197,18 @@ class gPodderPreferences(BuilderWidget):
 
         self.audio_player_model = self.user_apps_reader.get_model('audio')
         self.combo_audio_player_app.set_model(self.audio_player_model)
-        index = self.audio_player_model.get_index(self._config.player)
+        index = self.audio_player_model.get_index(self._config.player.audio)
         self.combo_audio_player_app.set_active(index)
 
         self.video_player_model = self.user_apps_reader.get_model('video')
         self.combo_video_player_app.set_model(self.video_player_model)
-        index = self.video_player_model.get_index(self._config.videoplayer)
+        index = self.video_player_model.get_index(self._config.player.video)
         self.combo_video_player_app.set_active(index)
 
         self.preferred_youtube_format_model = YouTubeVideoFormatListModel(self._config)
         self.combobox_preferred_youtube_format.set_model(self.preferred_youtube_format_model)
         cellrenderer = Gtk.CellRendererText()
+        cellrenderer.set_property('ellipsize', Pango.EllipsizeMode.END)
         self.combobox_preferred_youtube_format.pack_start(cellrenderer, True)
         self.combobox_preferred_youtube_format.add_attribute(cellrenderer, 'text', self.preferred_youtube_format_model.C_CAPTION)
         self.combobox_preferred_youtube_format.set_active(self.preferred_youtube_format_model.get_index())
@@ -214,6 +216,7 @@ class gPodderPreferences(BuilderWidget):
         self.preferred_youtube_hls_format_model = YouTubeVideoHLSFormatListModel(self._config)
         self.combobox_preferred_youtube_hls_format.set_model(self.preferred_youtube_hls_format_model)
         cellrenderer = Gtk.CellRendererText()
+        cellrenderer.set_property('ellipsize', Pango.EllipsizeMode.END)
         self.combobox_preferred_youtube_hls_format.pack_start(cellrenderer, True)
         self.combobox_preferred_youtube_hls_format.add_attribute(cellrenderer, 'text', self.preferred_youtube_hls_format_model.C_CAPTION)
         self.combobox_preferred_youtube_hls_format.set_active(self.preferred_youtube_hls_format_model.get_index())
@@ -221,31 +224,33 @@ class gPodderPreferences(BuilderWidget):
         self.preferred_vimeo_format_model = VimeoVideoFormatListModel(self._config)
         self.combobox_preferred_vimeo_format.set_model(self.preferred_vimeo_format_model)
         cellrenderer = Gtk.CellRendererText()
+        cellrenderer.set_property('ellipsize', Pango.EllipsizeMode.END)
         self.combobox_preferred_vimeo_format.pack_start(cellrenderer, True)
         self.combobox_preferred_vimeo_format.add_attribute(cellrenderer, 'text', self.preferred_vimeo_format_model.C_CAPTION)
         self.combobox_preferred_vimeo_format.set_active(self.preferred_vimeo_format_model.get_index())
 
-        self._config.connect_gtk_togglebutton('podcast_list_view_all',
-                                              self.checkbutton_show_all_episodes)
-        self._config.connect_gtk_togglebutton('podcast_list_sections',
-                                              self.checkbutton_podcast_sections)
+        self._config.connect_gtk_togglebutton('ui.gtk.find_as_you_type',
+                                              self.checkbutton_find_as_you_type)
 
         self.update_interval_presets = [0, 10, 30, 60, 2 * 60, 6 * 60, 12 * 60]
         adjustment_update_interval = self.hscale_update_interval.get_adjustment()
         adjustment_update_interval.set_upper(len(self.update_interval_presets) - 1)
-        if self._config.auto_update_frequency in self.update_interval_presets:
-            index = self.update_interval_presets.index(self._config.auto_update_frequency)
+        if self._config.auto.update.frequency in self.update_interval_presets:
+            index = self.update_interval_presets.index(self._config.auto.update.frequency)
             self.hscale_update_interval.set_value(index)
         else:
             # Patch in the current "custom" value into the mix
-            self.update_interval_presets.append(self._config.auto_update_frequency)
+            self.update_interval_presets.append(self._config.auto.update.frequency)
             self.update_interval_presets.sort()
 
             adjustment_update_interval.set_upper(len(self.update_interval_presets) - 1)
-            index = self.update_interval_presets.index(self._config.auto_update_frequency)
+            index = self.update_interval_presets.index(self._config.auto.update.frequency)
             self.hscale_update_interval.set_value(index)
 
-        self._config.connect_gtk_spinbutton('max_episodes_per_feed', self.spinbutton_episode_limit)
+        self._config.connect_gtk_spinbutton('limit.episodes', self.spinbutton_episode_limit)
+
+        self._config.connect_gtk_togglebutton('ui.gtk.only_added_are_new',
+                                              self.checkbutton_only_added_are_new)
 
         self.auto_download_model = NewEpisodeActionList(self._config)
         self.combo_auto_download.set_model(self.auto_download_model)
@@ -254,19 +259,22 @@ class gPodderPreferences(BuilderWidget):
         self.combo_auto_download.add_attribute(cellrenderer, 'text', NewEpisodeActionList.C_CAPTION)
         self.combo_auto_download.set_active(self.auto_download_model.get_index())
 
-        if self._config.auto_remove_played_episodes:
-            adjustment_expiration = self.hscale_expiration.get_adjustment()
-            if self._config.episode_old_age > adjustment_expiration.get_upper():
-                # Patch the adjustment to include the higher current value
-                adjustment_expiration.set_upper(self._config.episode_old_age)
+        self._config.connect_gtk_togglebutton('check_connection',
+                                              self.checkbutton_check_connection)
 
-            self.hscale_expiration.set_value(self._config.episode_old_age)
+        if self._config.auto.cleanup.played:
+            adjustment_expiration = self.hscale_expiration.get_adjustment()
+            if self._config.auto.cleanup.days > adjustment_expiration.get_upper():
+                # Patch the adjustment to include the higher current value
+                adjustment_expiration.set_upper(self._config.auto.cleanup.days)
+
+            self.hscale_expiration.set_value(self._config.auto.cleanup.days)
         else:
             self.hscale_expiration.set_value(0)
 
-        self._config.connect_gtk_togglebutton('auto_remove_unplayed_episodes',
+        self._config.connect_gtk_togglebutton('auto.cleanup.unplayed',
                                               self.checkbutton_expiration_unplayed)
-        self._config.connect_gtk_togglebutton('auto_remove_unfinished_episodes',
+        self._config.connect_gtk_togglebutton('auto.cleanup.unfinished',
                                               self.checkbutton_expiration_unfinished)
 
         self.device_type_model = DeviceTypeActionList(self._config)
@@ -290,6 +298,8 @@ class gPodderPreferences(BuilderWidget):
                                               self.checkbutton_create_playlists)
         self._config.connect_gtk_togglebutton('device_sync.playlists.two_way_sync',
                                               self.checkbutton_delete_using_playlists)
+        self._config.connect_gtk_togglebutton('device_sync.delete_deleted_episodes',
+                                              self.checkbutton_delete_deleted_episodes)
 
         # Have to do this before calling set_active on checkbutton_enable
         self._enable_mygpo = self._config.mygpo.enabled
@@ -311,10 +321,35 @@ class gPodderPreferences(BuilderWidget):
 
         gpodder.user_extensions.on_ui_object_available('preferences-gtk', self)
 
+        self.inject_extensions_preferences(init=True)
+
+        self.prefs_stack.foreach(self._wrap_checkbox_labels)
+
+    def _wrap_checkbox_labels(self, w, *args):
+        if w.get_name().startswith("no_label_wrap"):
+            return
+        elif isinstance(w, Gtk.CheckButton):
+            label = w.get_child()
+            label.set_line_wrap(True)
+        elif isinstance(w, Gtk.Container):
+            w.foreach(self._wrap_checkbox_labels)
+
+    def inject_extensions_preferences(self, init=False):
+        if not init:
+            # remove preferences buttons for all extensions
+            for child in self.prefs_stack.get_children():
+                if child.get_name().startswith("extension."):
+                    self.prefs_stack.remove(child)
+
+        # add preferences buttons for all extensions
         result = gpodder.user_extensions.on_preferences()
         if result:
             for label, callback in result:
-                self.notebook.append_page(callback(), Gtk.Label(label))
+                page = callback()
+                name = "extension." + label
+                page.set_name(name)
+                page.foreach(self._wrap_checkbox_labels)
+                self.prefs_stack.add_titled(page, name, label)
 
     def _extensions_select_function(self, selection, model, path, path_currently_selected):
         return model.get_value(model.get_iter(path), self.C_SHOW_TOGGLE)
@@ -443,6 +478,7 @@ class gPodderPreferences(BuilderWidget):
                 self.on_extension_enabled(container.module)
             else:
                 self.on_extension_disabled(container.module)
+            self.inject_extensions_preferences()
         elif container.error is not None:
             if hasattr(container.error, 'message'):
                 error_msg = container.error.message
@@ -456,13 +492,11 @@ class gPodderPreferences(BuilderWidget):
         if not container or not model:
             return
 
-        # This is one ugly hack, but it displays the attributes of
-        # the metadata object of the container..
-        info = '\n'.join('<b>%s:</b> %s' %
-                         tuple(map(html.escape, list(map(str, (key, value)))))
-                         for key, value in container.metadata.get_sorted())
+        info = '\n'.join('<b>{}:</b> {}'.format(html.escape(key), html.escape(value))
+                         for key, value in container.metadata.get_sorted()
+                         if key not in ('title', 'description'))
 
-        self.show_message(info, _('Extension module info'), important=True)
+        self.show_message_details(container.metadata.title, container.metadata.description, info)
 
     def open_weblink(self, w, url):
         util.open_website(url)
@@ -484,11 +518,11 @@ class gPodderPreferences(BuilderWidget):
 
     def on_combo_audio_player_app_changed(self, widget):
         index = self.combo_audio_player_app.get_active()
-        self._config.player = self.audio_player_model.get_command(index)
+        self._config.player.audio = self.audio_player_model.get_command(index)
 
     def on_combo_video_player_app_changed(self, widget):
         index = self.combo_video_player_app.get_active()
-        self._config.videoplayer = self.video_player_model.get_command(index)
+        self._config.player.video = self.video_player_model.get_command(index)
 
     def on_combobox_preferred_youtube_format_changed(self, widget):
         index = self.combobox_preferred_youtube_format.get_active()
@@ -505,21 +539,21 @@ class gPodderPreferences(BuilderWidget):
     def on_button_audio_player_clicked(self, widget):
         result = self.show_text_edit_dialog(_('Configure audio player'),
                                             _('Command:'),
-                                            self._config.player)
+                                            self._config.player.audio)
 
         if result:
-            self._config.player = result
-            index = self.audio_player_model.get_index(self._config.player)
+            self._config.player.audio = result
+            index = self.audio_player_model.get_index(self._config.player.audio)
             self.combo_audio_player_app.set_active(index)
 
     def on_button_video_player_clicked(self, widget):
         result = self.show_text_edit_dialog(_('Configure video player'),
                                             _('Command:'),
-                                            self._config.videoplayer)
+                                            self._config.player.video)
 
         if result:
-            self._config.videoplayer = result
-            index = self.video_player_model.get_index(self._config.videoplayer)
+            self._config.player.video = result
+            index = self.video_player_model.get_index(self._config.player.video)
             self.combo_video_player_app.set_active(index)
 
     def format_update_interval_value(self, scale, value):
@@ -540,8 +574,8 @@ class gPodderPreferences(BuilderWidget):
 
     def on_update_interval_value_changed(self, range):
         value = int(range.get_value())
-        self._config.auto_update_feeds = (value > 0)
-        self._config.auto_update_frequency = self.update_interval_presets[value]
+        self._config.auto.update.enabled = (value > 0)
+        self._config.auto.update.frequency = self.update_interval_presets[value]
 
     def on_combo_auto_download_changed(self, widget):
         index = self.combo_auto_download.get_active()
@@ -560,11 +594,11 @@ class gPodderPreferences(BuilderWidget):
 
         if value == 0:
             self.checkbutton_expiration_unplayed.set_active(False)
-            self._config.auto_remove_played_episodes = False
-            self._config.auto_remove_unplayed_episodes = False
+            self._config.auto.cleanup.played = False
+            self._config.auto.cleanup.unplayed = False
         else:
-            self._config.auto_remove_played_episodes = True
-            self._config.episode_old_age = value
+            self._config.auto.cleanup.played = True
+            self._config.auto.cleanup.days = value
 
         self.checkbutton_expiration_unplayed.set_sensitive(value > 0)
         self.checkbutton_expiration_unfinished.set_sensitive(value > 0)
@@ -618,7 +652,8 @@ class gPodderPreferences(BuilderWidget):
             children = self.btn_playlistfolder.get_children()
             if children:
                 label = children.pop()
-                label.set_alignment(0., .5)
+                label.set_ellipsize(Pango.EllipsizeMode.START)
+                label.set_xalign(0.0)
         else:
             self.btn_playlistfolder.set_sensitive(False)
             self.btn_playlistfolder.set_label('')
@@ -637,16 +672,13 @@ class gPodderPreferences(BuilderWidget):
             self.combobox_on_sync.set_sensitive(False)
             self.checkbutton_skip_played_episodes.set_sensitive(False)
         elif device_type == 'filesystem':
-            self.btn_filesystemMountpoint.set_label(self._config.device_sync.device_folder)
+            self.btn_filesystemMountpoint.set_label(self._config.device_sync.device_folder or "")
             self.btn_filesystemMountpoint.set_sensitive(True)
             self.checkbutton_create_playlists.set_sensitive(True)
-            children = self.btn_filesystemMountpoint.get_children()
-            if children:
-                label = children.pop()
-                label.set_alignment(0., .5)
             self.toggle_playlist_interface(self._config.device_sync.playlists.create)
             self.combobox_on_sync.set_sensitive(True)
             self.checkbutton_skip_played_episodes.set_sensitive(True)
+            self.checkbutton_delete_deleted_episodes.set_sensitive(True)
         elif device_type == 'ipod':
             self.btn_filesystemMountpoint.set_label(self._config.device_sync.device_folder)
             self.btn_filesystemMountpoint.set_sensitive(True)
@@ -654,29 +686,28 @@ class gPodderPreferences(BuilderWidget):
             self.toggle_playlist_interface(False)
             self.checkbutton_delete_using_playlists.set_sensitive(False)
             self.combobox_on_sync.set_sensitive(False)
-            self.checkbutton_skip_played_episodes.set_sensitive(False)
+            self.checkbutton_skip_played_episodes.set_sensitive(True)
+            self.checkbutton_delete_deleted_episodes.set_sensitive(True)
 
-            children = self.btn_filesystemMountpoint.get_children()
-            if children:
-                label = children.pop()
-                label.set_alignment(0., .5)
-
-        else:
-            # TODO: Add support for iPod and MTP devices
-            pass
+        children = self.btn_filesystemMountpoint.get_children()
+        if children:
+            label = children.pop()
+            label.set_ellipsize(Pango.EllipsizeMode.START)
+            label.set_xalign(0.0)
 
     def on_btn_device_mountpoint_clicked(self, widget):
         fs = Gtk.FileChooserDialog(title=_('Select folder for mount point'),
                 action=Gtk.FileChooserAction.SELECT_FOLDER)
-        fs.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
-        fs.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
-        fs.set_current_folder(self.btn_filesystemMountpoint.get_label())
+        fs.set_local_only(False)
+        fs.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL)
+        fs.add_button(_('_Open'), Gtk.ResponseType.OK)
+
+        fs.set_uri(self.btn_filesystemMountpoint.get_label() or "")
         if fs.run() == Gtk.ResponseType.OK:
-            filename = fs.get_filename()
             if self._config.device_sync.device_type == 'filesystem':
-                self._config.device_sync.device_folder = filename
+                self._config.device_sync.device_folder = fs.get_uri()
             elif self._config.device_sync.device_type == 'ipod':
-                self._config.device_sync.device_folder = filename
+                self._config.device_sync.device_folder = fs.get_filename()
             # Request an update of the mountpoint button
             self.on_combobox_device_type_changed(None)
 
@@ -685,18 +716,29 @@ class gPodderPreferences(BuilderWidget):
     def on_btn_playlist_folder_clicked(self, widget):
         fs = Gtk.FileChooserDialog(title=_('Select folder for playlists'),
                 action=Gtk.FileChooserAction.SELECT_FOLDER)
-        fs.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
-        fs.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
-        fs.set_current_folder(self.btn_playlistfolder.get_label())
-        if fs.run() == Gtk.ResponseType.OK:
-            filename = util.relpath(self._config.device_sync.device_folder,
-                                    fs.get_filename())
+        fs.set_local_only(False)
+        fs.add_button(_('_Cancel'), Gtk.ResponseType.CANCEL)
+        fs.add_button(_('_Open'), Gtk.ResponseType.OK)
+
+        device_folder = util.new_gio_file(self._config.device_sync.device_folder)
+        playlists_folder = device_folder.resolve_relative_path(self._config.device_sync.playlists.folder)
+        fs.set_file(playlists_folder)
+
+        while fs.run() == Gtk.ResponseType.OK:
+            filename = util.relpath(fs.get_uri(),
+                                    self._config.device_sync.device_folder)
+            if not filename:
+                show_message_dialog(fs, _('The playlists folder must be on the device'))
+                continue
+
             if self._config.device_sync.device_type == 'filesystem':
                 self._config.device_sync.playlists.folder = filename
-                self.btn_playlistfolder.set_label(filename)
+                self.btn_playlistfolder.set_label(filename or "")
                 children = self.btn_playlistfolder.get_children()
                 if children:
                     label = children.pop()
-                    label.set_alignment(0., .5)
+                    label.set_ellipsize(Pango.EllipsizeMode.START)
+                    label.set_xalign(0.0)
+            break
 
         fs.destroy()
