@@ -18,7 +18,6 @@
 #
 import html
 import logging
-import re
 from urllib.parse import urlparse
 
 import gpodder
@@ -31,7 +30,7 @@ from gpodder.gtkui.draw import (draw_text_box_centered, get_background_color,
 import gi  # isort:skip
 gi.require_version('Gdk', '3.0')  # isort:skip
 gi.require_version('Gtk', '3.0')  # isort:skip
-from gi.repository import Gdk, Gtk, Pango  # isort:skip
+from gi.repository import Gdk, Gio, GLib, Gtk, Pango  # isort:skip
 
 
 _ = gpodder.gettext
@@ -193,9 +192,12 @@ class gPodderShownotesText(gPodderShownotes):
         self.text_view.connect('button-release-event', self.on_button_release)
         self.text_view.connect('key-press-event', self.on_key_press)
         self.text_view.connect('motion-notify-event', self.on_hover_hyperlink)
+        self.populate_popup_id = None
         return self.text_view
 
     def update(self, episode):
+        self.scrolled_window.get_vadjustment().set_value(0)
+
         heading = episode.title
         subheading = _('from %s') % (episode.channel.title)
         details = self.details_fmt % {
@@ -206,13 +208,17 @@ class gPodderShownotesText(gPodderShownotes):
         self.define_colors()
         hyperlinks = [(0, None)]
         self.text_buffer.set_text('')
+        if episode.link:
+            hyperlinks.append((self.text_buffer.get_char_count(), episode.link))
         self.text_buffer.insert_with_tags_by_name(self.text_buffer.get_end_iter(), heading, 'heading')
+        if episode.link:
+            hyperlinks.append((self.text_buffer.get_char_count(), None))
         self.text_buffer.insert_at_cursor('\n')
         self.text_buffer.insert_with_tags_by_name(self.text_buffer.get_end_iter(), subheading, 'subheading')
         self.text_buffer.insert_at_cursor('\n')
         self.text_buffer.insert_with_tags_by_name(self.text_buffer.get_end_iter(), details, 'details')
         self.text_buffer.insert_at_cursor('\n\n')
-        for target, text in util.extract_hyperlinked_text(episode.description_html or episode.description):
+        for target, text in util.extract_hyperlinked_text(episode.html_description()):
             hyperlinks.append((self.text_buffer.get_char_count(), target))
             if target:
                 self.text_buffer.insert_with_tags_by_name(
@@ -224,6 +230,34 @@ class gPodderShownotesText(gPodderShownotes):
         self.hyperlinks = [(start, end, url) for (start, url), (end, _) in zip(hyperlinks, hyperlinks[1:]) if url]
         self.text_buffer.place_cursor(self.text_buffer.get_start_iter())
 
+        if self.populate_popup_id is not None:
+            self.text_view.disconnect(self.populate_popup_id)
+        self.populate_popup_id = self.text_view.connect('populate-popup', self.on_populate_popup)
+        self.episode = episode
+
+    def on_populate_popup(self, textview, context_menu):
+        # TODO: Remove items from context menu that are always insensitive in a read-only buffer
+
+        if self.episode.link:
+            # TODO: It is currently not possible to copy links in description.
+            # Detect if context menu was opened on a hyperlink and add
+            # "Open Link" and "Copy Link Address" menu items.
+            # See https://github.com/gpodder/gpodder/issues/1097
+
+            item = Gtk.SeparatorMenuItem()
+            item.show()
+            context_menu.append(item)
+            # label links can be opened from context menu or by clicking them, do the same here
+            item = Gtk.MenuItem(label=_('Open Episode Title Link'))
+            item.connect('activate', lambda i: util.open_website(self.episode.link))
+            item.show()
+            context_menu.append(item)
+            # hack to allow copying episode.link
+            item = Gtk.MenuItem(label=_('Copy Episode Title Link Address'))
+            item.connect('activate', lambda i: util.copy_text_to_clipboard(self.episode.link))
+            item.show()
+            context_menu.append(item)
+
     def on_button_release(self, widget, event):
         if event.button == 1:
             self.activate_links()
@@ -287,6 +321,8 @@ class gPodderShownotesHTML(gPodderShownotes):
         return self.html_view
 
     def update(self, episode):
+        self.scrolled_window.get_vadjustment().set_value(0)
+
         self.define_colors()
 
         if episode.has_website_link():
@@ -312,13 +348,10 @@ class gPodderShownotesHTML(gPodderShownotes):
             'duration': episode.get_play_info_string()})
         header_html = _('<div id="gpodder-title">\n%(heading)s\n<p>%(subheading)s</p>\n<p>%(details)s</p></div>\n') \
             % dict(heading=heading, subheading=subheading, details=details)
-        description_html = episode.description_html
-        if not description_html:
-            description_html = re.sub(r'\n', '<br>\n', episode.description)
         # uncomment to prevent background override in html shownotes
         # self.manager.remove_all_style_sheets ()
         logger.debug("base uri: %s (chan:%s)", self._base_uri, episode.channel.url)
-        self.html_view.load_html(header_html + description_html, self._base_uri)
+        self.html_view.load_html(header_html + episode.html_description(), self._base_uri)
         # uncomment to show web inspector
         # self.html_view.get_inspector().show()
         self.episode = episode
@@ -374,15 +407,16 @@ class gPodderShownotesHTML(gPodderShownotes):
             if req.get_uri() in (self._base_uri, 'about:blank'):
                 decision.use()
             else:
-                logger.debug("refusing to go to %s (base URI=%s)", req.get_uri(), self._base_uri)
+                # Avoid opening the page inside the WebView and open in the browser instead
                 decision.ignore()
+                util.open_website(req.get_uri())
             return False
         else:
             decision.use()
             return False
 
-    def on_open_in_browser(self, action):
-        util.open_website(action.url)
+    def on_open_in_browser(self, action, var):
+        util.open_website(var.get_string())
 
     def on_authenticate(self, view, request):
         if request.is_retry():
@@ -412,10 +446,10 @@ class gPodderShownotesHTML(gPodderShownotes):
             return False
 
     def create_open_item(self, name, label, url):
-        action = Gtk.Action.new(name, label, None, Gtk.STOCK_OPEN)
-        action.url = url
+        action = Gio.SimpleAction.new(name, GLib.VariantType.new('s'))
         action.connect('activate', self.on_open_in_browser)
-        return WebKit2.ContextMenuItem.new(action)
+        var = GLib.Variant.new_string(url)
+        return WebKit2.ContextMenuItem.new_from_gaction(action, label, var)
 
     def get_stylesheet(self):
         if self.stylesheet is None:
