@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # -*- coding: utf-8 -*-
 #
 # gPodder - A media aggregator and podcast client
@@ -255,8 +254,8 @@ class MygPoClient(object):
                 logger.debug('Play action for %s', episode.url)
                 episode.mark(is_played=True)
 
-                if (action.timestamp > episode.current_position_updated and
-                        action.position is not None):
+                if (action.timestamp > episode.current_position_updated
+                        and action.position is not None):
                     logger.debug('Updating position for %s', episode.url)
                     episode.current_position = action.position
                     episode.current_position_updated = action.timestamp
@@ -445,7 +444,7 @@ class MygPoClient(object):
 
     def flush(self, now=False):
         if not self.can_access_webservice():
-            logger.warn('Flush requested, but sync disabled.')
+            logger.warning('Flush requested, but sync disabled.')
             return
 
         if self._worker_thread is None or now:
@@ -512,7 +511,7 @@ class MygPoClient(object):
                 raise
 
             except Exception as e:
-                logger.warn('Exception while polling for episodes.', exc_info=True)
+                logger.warning('Exception while polling for episodes.', exc_info=True)
 
             # Step 2: Upload Episode actions
 
@@ -534,7 +533,7 @@ class MygPoClient(object):
             return True
 
         except (MissingCredentials, mygpoclient.http.Unauthorized):
-            logger.warn('Invalid credentials. Disabling gpodder.net.')
+            logger.warning('Invalid credentials. Disabling gpodder.net.')
             self._config.mygpo.enabled = False
             return False
 
@@ -599,7 +598,7 @@ class MygPoClient(object):
             return True
 
         except (MissingCredentials, mygpoclient.http.Unauthorized):
-            logger.warn('Invalid credentials. Disabling gpodder.net.')
+            logger.warning('Invalid credentials. Disabling gpodder.net.')
             self._config.mygpo.enabled = False
             return False
 
@@ -616,7 +615,7 @@ class MygPoClient(object):
             return True
 
         except (MissingCredentials, mygpoclient.http.Unauthorized):
-            logger.warn('Invalid credentials. Disabling gpodder.net.')
+            logger.warning('Invalid credentials. Disabling gpodder.net.')
             self._config.mygpo.enabled = False
             return False
 
@@ -632,7 +631,7 @@ class MygPoClient(object):
             devices = self._client.get_devices()
 
         except (MissingCredentials, mygpoclient.http.Unauthorized):
-            logger.warn('Invalid credentials. Disabling gpodder.net.')
+            logger.warning('Invalid credentials. Disabling gpodder.net.')
             self._config.mygpo.enabled = False
             raise
 
